#include "jserver.h"

void js_presence_incoming(udata user, jpacket p)
{
    xmlnode item, reply;

    jdebug(ZONE,"incoming presence %s request for %s from %s",xmlnode_get_attrib(p->x,"type"),user->user,jid_full(p->from));

    item = js_roster_item(user,p->from);

    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"subscribe") == 0)
    {
        /* cache the subscription for later reference */
        if(item == NULL)
        {
            item = js_roster_new(user,p->from);
            xmlnode_put_attrib(item,"server","yes");
        }
        xmlnode_insert_tag(item,"subscribe");
    }
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"subscribed") == 0)
    {
        xmlnode_hide_attrib(item,"ask");
        if(item == NULL)
        {
            item = js_roster_new(user,p->from);
            xmlnode_put_attrib(item,"server","yes");
        }

        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"from") == 0)
            xmlnode_put_attrib(item,"subscription","both");
        else
            xmlnode_put_attrib(item,"subscription","to");
        js_roster_push(user,item,NULL);
    }
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unsubscribe") == 0)
    {
        /* auto-unsubscribe reply */
        reply = xmlnode_new_tag("presence");
        xmlnode_put_attrib(reply,"type","unsubscribed");
        xmlnode_put_attrib(reply,"to",jid_full(p->from));
        jid_set(p->to,NULL,JID_RESOURCE);
        xmlnode_put_attrib(reply,"from",jid_full(p->to));
        deliver(jpacket_new(reply));

        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0)
            xmlnode_put_attrib(item,"subscription","to");
        else
            xmlnode_put_attrib(item,"subscription","none");
        js_roster_push(user,item,NULL);
    }
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unsubscribed") == 0)
    {
        xmlnode_hide_attrib(item,"ask");
        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0)
            xmlnode_put_attrib(item,"subscription","from");
        else
            xmlnode_put_attrib(item,"subscription","none");
        js_roster_push(user,item,NULL);
    }

}

void js_presence_outgoing(session s, jpacket p)
{
    xmlnode item, reply;

    jdebug(ZONE,"outgoing presence %s request for %s from %s",xmlnode_get_attrib(p->x,"type"),jid_full(p->to),s->u->user);

    item = js_roster_item(s->u,p->to);

    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"subscribe") == 0)
    {
        if(item == NULL)
        {
            item = js_roster_new(s->u,p->to);
            xmlnode_put_attrib(item,"server","yes");
        }
        xmlnode_put_attrib(item,"ask","subscribe");
        js_roster_push(s->u,item,NULL);
    }
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"subscribed") == 0)
    {
        if(item == NULL)
        {
            item = js_roster_new(s->u,p->to);
            xmlnode_put_attrib(item,"server","yes");
        }
        xmlnode_hide(xmlnode_get_tag(item,"subscribe")); /* clear cache */
        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"to") == 0)
            xmlnode_put_attrib(item,"subscription","both");
        else
            xmlnode_put_attrib(item,"subscription","from");
        js_roster_push(s->u,item,NULL);

        /* now they need to know about us, so send a probe to ourselves from them :) */
        item = xmlnode_new_tag("presence");
        xmlnode_put_attrib(item,"type","probe");
        xmlnode_put_attrib(item,"to",jid_full(p->from));
        xmlnode_put_attrib(item,"from",jid_full(p->to));
        deliver(jpacket_new(item));
    }
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unsubscribe") == 0)
    {
        xmlnode_put_attrib(item,"ask","unsubscribe");
        js_roster_push(s->u,item,NULL);
    }
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unsubscribed") == 0)
    {
        xmlnode_hide(xmlnode_get_tag(item,"subscribe")); /* clear cache */
        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0)
            xmlnode_put_attrib(item,"subscription","to");
        else
            xmlnode_put_attrib(item,"subscription","none");
        js_roster_push(s->u,item,NULL);
    }

}

jid js_notification(int cmd, session s, jid to)
{
    jid curr;

    jdebug(ZONE,"notification cache %d for %s to %s",cmd,jid_full(s->id),jid_full(to));

    switch(cmd)
    {
        case CMD_SET:
            s->notification = jid_append(s->notification, jid_new(s->p,jid_full(to)));
            break;
        case CMD_GET:
            for(curr = s->notification;curr != NULL && jid_cmp(curr,to) != 0;curr = curr->next);
            return curr;
            break;
        case CMD_DEL:
            if(jid_cmp(to,s->notification) == 0)
            {
                s->notification = s->notification->next;
            }else{
                for(curr = s->notification;curr != NULL && jid_cmp(curr->next,to) != 0;curr = curr->next);
                if(curr != NULL && curr->next != NULL)
                    curr->next = curr->next->next;
            }
            break;
    }

    return NULL;
}

/* send the presence packet out */
void js_presence_handle(session s, jpacket p)
{
    char *group = NULL;
    int force_deliver, priority = js_priority(p->x);
    xmlnode item, grp, newp, newp2;
    jid uid;
    jpacket jpnew;

    jdebug(ZONE,"presence type=%s to %s from %s priority %d",xmlnode_get_attrib(p->x,"type"),jid_full(p->to),jid_full(p->from),priority);

    if(p->to == NULL)
    {
        if(priority < 0 && j_strcmp(xmlnode_get_attrib(p->x,"type"),"unavailable") != 0)
            return; /* bum packet, croak */

        s->presence = xmlnode_dup(p->x); /* save this as the main presence for probes */

        /* since this is master, also force send to ourselves */
        newp = xmlnode_dup(p->x);
        uid = jid_new(xmlnode_pool(newp),jid_full(s->id));
        jid_set(uid,NULL,JID_RESOURCE);
        xmlnode_put_attrib(newp,"to",jid_full(uid));
        /* so that our other resources get notified */
        if(priority >= 0 && js_notification(CMD_GET,s,p->to) == NULL)
        {
            js_notification(CMD_SET,s,uid);
            /* tell ourselves about our other resources */
            while((newp2 = ppdb_get(s->u->cache,uid)) != NULL)
            {
                jpnew = jpacket_new(xmlnode_dup(newp2));
                if(jid_cmp(s->id,jpnew->from) != 0)
                    (s->send_evt)(s->t, jpnew);
            }
        }
        if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unavailable") == 0)
            js_notification(CMD_DEL,s,uid);
        /* reset primary */
        s->r->priority = priority;
        js_resetprimary(s->u);
        deliver(jpacket_new(newp));
    }else if(j_strcmp(p->to->type,"group") == 0){
        group = p->to->server;
    }else{
        /* deliver normally to the to */
        if(priority >= 0)
        {
            js_notification(CMD_SET,s,p->to);
        }else if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unavailable") == 0){
            js_notification(CMD_DEL,s,p->to);
        }else{
            /* sending out a [un]subscribe[d], make sure it's from the user and not the resouce */
            jid_set(p->from,NULL,JID_RESOURCE);
            xmlnode_put_attrib(p->x,"from",jid_full(p->from));
            js_presence_outgoing(s,p);
        }
        if(j_strcmp(p->to->user,p->from->user) == 0 && j_strcmp(p->to->server,p->from->server) == 0)
        {/* to ourselves, so this is "our" presence, adjust defaulting accordingly */
            s->r->priority = priority;
            js_resetprimary(s->u);
        }
        deliver(p);
        return;
    }

    jdebug(ZONE,"presence delivery from roster");

    /* loop through the roster copying to all matching items */
    for(item = xmlnode_get_firstchild(js_roster(s->u)); item != NULL; item = xmlnode_get_nextsibling(item))
    {
        if(group != NULL)
        {
            for(grp = xmlnode_get_firstchild(item); grp != NULL; grp = xmlnode_get_nextsibling(grp))
                if(j_strcmp(xmlnode_get_name(grp),"group") == 0 && j_strcmp(xmlnode_get_data(xmlnode_get_firstchild(grp)),group) == 0)
                    break;
            if(grp == NULL)
                break;
        }

        jdebug(ZONE,"roster item %s subscription %s",xmlnode_get_attrib(item,"jid"),xmlnode_get_attrib(item,"subscription"));

        xmlnode_put_attrib(p->x,"to",xmlnode_get_attrib(item,"jid"));
        p->to = jid_new(p->p,xmlnode_get_attrib(item,"jid"));

        if(priority >= 0 && js_notification(CMD_GET,s,p->to) == NULL)
        { /* we're telling them we're available for the first */
            if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"to") == 0 || j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0)
            { /* we are subscribed to their presence */
                if(ppdb_primary(s->u->cache,p->to) == NULL)
                { /* either they've never told us their presence, or they arent available, ask them to be sure */
                    jdebug(ZONE,"probe request generated");
                    newp = xmlnode_new_tag("presence");
                    xmlnode_put_attrib(newp,"type","probe");
                    xmlnode_put_attrib(newp,"to",xmlnode_get_attrib(item,"jid"));
                    uid = jid_new(xmlnode_pool(newp),jid_full(s->id));
                    jid_set(uid,NULL,JID_RESOURCE);
                    xmlnode_put_attrib(newp,"from",jid_full(uid));
                    deliver(jpacket_new(newp));
                    force_deliver = 1;
                }else{
                    /* go through the cache from this guy and feed it to the connection */
                    jdebug(ZONE,"use cache and feed to the client");
                    while((newp = ppdb_get(s->u->cache,p->to)) != NULL)
                    {
                        (s->send_evt)(s->t, jpacket_new(xmlnode_dup(newp)));
                    }
                }
            }
            js_notification(CMD_SET,s,p->to);
        }

        if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"unavailable") == 0)
            js_notification(CMD_DEL,s,jid_new(p->p,xmlnode_get_attrib(item,"jid")));

        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"from") == 0)
            deliver(jpacket_new(xmlnode_dup(p->x)));
        if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0 && (ppdb_primary(s->u->cache,p->to) != NULL || force_deliver == 1))
        { /* if they are subscribed to us, we can be efficient and only send it to them if they are available, or this is the first time */
            deliver(jpacket_new(xmlnode_dup(p->x)));
            force_deliver = 0;
        }else{
            jdebug(ZONE,"not online or no subscription, don't forward");
        }
    }

    xmlnode_free(p->x);
}

