#include "jserver.h"

xmlnode js_roster(udata user)
{
    pool p;
    mapi m;

    if(user == NULL)
        return NULL;

    jdebug(ZONE,"js_roster %s",user->user);

    if(user->roster == NULL)
    {
        p = pool_new(); /* temporary pool for this module call */
        m = mapi_new(M_PHASE_ROSTER, p, NULL, NULL);
        m->cmd = CMD_INIT;
        m->user = pstrdup(p,user->user);
        m->roster = NULL;
        module_dispatch(m); /* let the module initialize the user's roster */
        user->roster = m->roster;
        pool_free(p);
    }

    return user->roster;
}

void js_roster_push(udata user, xmlnode item, resource ignore)
{
    resource res;
    xmlnode iq, query, i2;
    session s;
    mapi m;
    pool p;

    if(item == NULL)
        return;

    jdebug(ZONE,"pushing %s roster item %s",user->user,xmlnode2str(item));

    /* push the roster item to all the sessions that have a roster for this user */
    for(res = user->resources; res != NULL; res = res->next)
    {
        if(res->type == RESOURCE_SESSION && res != ignore)
        {
            s = (session)res->aux;
            if(s->roster_flag == 0)
                continue;

            /* make an iq for the item and copy into it */
            iq = xmlnode_new_tag("iq");
            xmlnode_put_attrib(iq,"type","set");
            query = xmlnode_insert_tag(iq,"query");
            xmlnode_put_attrib(query,"xmlns","jabber:iq:roster");
            i2 = xmlnode_insert_tag(query,"item");
            xmlnode_insert_node(i2, xmlnode_get_firstattrib(item));
            xmlnode_insert_node(i2, xmlnode_get_firstchild(item));
            /* hide the server only stuff */
            xmlnode_hide_attrib(i2,"server");
            xmlnode_hide(xmlnode_get_tag(i2,"subscribe"));

            psend(res->handler,jpacket_new(iq));
        }
    }

    /* we're sending out a new roster item, let the module save the change to the roster */
    p = pool_new();
    m = mapi_new(M_PHASE_ROSTER, p, NULL, NULL);
    m->cmd = CMD_CHECK;
    m->user = pstrdup(p,user->user);
    m->roster = user->roster;
    module_dispatch(m); 
    pool_free(p);
}

xmlnode js_roster_item(udata user, jid id)
{
    xmlnode roster, item;

    roster = js_roster(user);

    jdebug(ZONE,"getting %s roster item %s",user->user,jid_full(id));
 
    for(item = xmlnode_get_firstchild(roster); item != NULL && j_strcmp(xmlnode_get_attrib(item,"jid"),jid_full(id)) != 0; item = xmlnode_get_nextsibling(item));

    return item;
}

xmlnode js_roster_new(udata user, jid id)
{
    xmlnode roster, item;

    roster = js_roster(user);

    jdebug(ZONE,"roster_new %s item %s",user->user,jid_full(id));

    /* find the to in the roster, fail, insert an item for it */
    item = xmlnode_insert_tag(roster,"item");
    xmlnode_put_attrib(item,"jid",jid_full(id));
    xmlnode_put_attrib(item,"subscription","none");
    xmlnode_put_attrib(item,"server","yes");

    return item;
}

void session_roster_handle(session s, jpacket p)
{
    xmlnode roster, curr, item, pres;
    mapi m;
    jid id, cache = NULL;
    
    roster = js_roster(s->u);

    jdebug(ZONE,"roster handler for %s",jid_full(s->id));

    s->roster_flag = 1;

    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"get") == 0)
    {
        for(curr = xmlnode_get_firstchild(roster); curr != NULL; curr = xmlnode_get_nextsibling(curr))
        {
            item = xmlnode_dup(curr);
            xmlnode_hide_attrib(item,"server");
            if(xmlnode_get_tag(item,"subscribe"))
            {
                cache = jid_append(cache,jid_new(p->p,xmlnode_get_attrib(item,"jid")));
                xmlnode_hide(xmlnode_get_tag(item,"subscribe"));
            }
            xmlnode_insert_node(p->iq,item);
            xmlnode_free(item);
        }
        js_iqreply(p->x);
        psend(s->mp_to, p);

        for(;cache != NULL; cache = cache->next)
        {
            pres = xmlnode_new_tag("presence");
            xmlnode_put_attrib(pres,"type","subscribe");
            xmlnode_put_attrib(pres,"from",jid_full(cache));
            psend(s->mp_to,jpacket_new(pres));
        }
        return;
    }

    id = jid_new(p->p,jid_full(s->id));
    jid_set(id,NULL,JID_RESOURCE);
    m = mapi_new(M_PHASE_ROSTER, p->p, s, p);
    m->cmd = CMD_SET;
    for(curr = xmlnode_get_firstchild(p->iq); curr != NULL; curr = xmlnode_get_nextsibling(curr))
    {
        xmlnode_hide(curr);
        if(xmlnode_get_type(curr) != NTYPE_TAG || xmlnode_get_attrib(curr,"jid") == NULL)
            continue;

        m->result = M_DEFAULT;
        module_dispatch(m);
        if(m->result != M_DEFAULT)
            continue;

        /* no module handled it, use internal roster */
        jdebug(ZONE,"roster update");

        /* zoom to find the existing item in the current roster */
        item = js_roster_item(s->u,jid_new(p->p,xmlnode_get_attrib(curr,"jid")));

        /* drop you sukkah */
        if(j_strcmp(xmlnode_get_attrib(curr,"subscription"),"remove") == 0)
        {
            xmlnode_hide(item);
            if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0 || j_strcmp(xmlnode_get_attrib(item,"subscription"),"to") == 0)
            {
                pres = xmlnode_new_tag("presence");
                xmlnode_put_attrib(pres,"type","unsubscribe");
                xmlnode_put_attrib(pres,"from",jid_full(id));
                deliver(jpacket_new(pres));
            }
            if(j_strcmp(xmlnode_get_attrib(item,"subscription"),"both") == 0 || j_strcmp(xmlnode_get_attrib(item,"subscription"),"from") == 0)
            {
                pres = xmlnode_new_tag("presence");
                xmlnode_put_attrib(pres,"type","unsubscribed");
                xmlnode_put_attrib(pres,"from",jid_full(id));
                deliver(jpacket_new(pres));
            }
            js_roster_push(s->u,curr,s->r);
            continue;
        }

        if(item == NULL)
            item = js_roster_new(s->u,jid_new(p->p,xmlnode_get_attrib(curr,"jid")));

        /* remove old groups */
        while(xmlnode_get_firstchild(item) != NULL)
            xmlnode_hide(xmlnode_get_firstchild(item));
        /* copy new groups */
        if(xmlnode_has_children(curr))
            xmlnode_insert_node(item, xmlnode_get_firstchild(curr));
        /* copy name attrib */
        xmlnode_put_attrib(item,"name",xmlnode_get_attrib(curr,"name"));
        /* push the change'd item to other connections, but not us */
        js_roster_push(s->u,item,s->r);

    }
    /* tell the module that the roster has lump-changed */

    jdebug(ZONE,"SROSTER: %s",xmlnode2str(roster));
    js_iqreply(p->x);
    psend(s->mp_to, p);
}

